/*======================================================================

    PCMCIA Card Manager

    Written by David Hinds, dhinds@allegro.stanford.edu
    
======================================================================*/

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <syslog.h>
#include <getopt.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <linux/fs.h>

#include "cs_types.h"
#include "cs.h"
#include "cistpl.h"
#include "ds.h"

#include "cardmgr.h"

static const char *version =
    "cardmgr.c $Revision: 1.22 $ $Date: 1994/09/01 06:55:13 $ (David Hinds)";

/*====================================================================*/

typedef struct socket_info_t {
    int fd;
    int state;
    card_info_t *card;
    bind_info_t bind[MAX_FUNCTIONS];
} socket_info_t;

#define SOCKET_PRESENT	0x01
#define SOCKET_READY	0x02
#define SOCKET_BOUND	0x04

/* Linked list of resource adjustments */
struct adjust_list_t *root_adjust = NULL;

/* Linked list of device definitions */
struct device_info_t *root_device = NULL;

/* Linked list of card definitions */
struct card_info_t *root_card = NULL;

static struct socket_info_t socket[MAX_SOCKS];

/* Default path for config file */
#ifdef ETC
static char *config = ETC "/config";
#else
static char *config = "/etc/pcmcia/config";
#endif

/* Default path for finding modules */
#ifdef MODDIR
static char *modpath = MODDIR;
#else
static char *modpath = "/usr/src/linux/modules";
#endif

/*====================================================================*/

int lookup_dev(char *name)
{
    FILE *f;
    int n;
    char s[32], t[32];
    
    f = fopen("/proc/devices", "r");
    if (f == NULL)
	return -1;
    while (fgets(s, 32, f) != NULL) {
	if (sscanf(s, "%d %s", &n, t) == 2)
	    if (strcmp(name, t) == 0)
		break;
    }
    fclose(f);
    if (strcmp(name, t) == 0)
	return n;
    else
	return -1;
} /* lookup_dev */

/*====================================================================*/

int open_dev(dev_t dev)
{
    static char *fn;
    int fd;

    if ((fn = tmpnam(NULL)) == NULL)
	return -1;
    if (mknod(fn, (S_IFCHR|S_IREAD|S_IWRITE), dev) != 0)
	return -1;
    fd = open(fn, O_RDWR);
    unlink(fn);
    return fd;
} /* open_dev */

/*====================================================================*/

card_info_t *lookup_card(int ns)
{
    socket_info_t *s = &socket[ns];
    card_info_t *card;
    ds_ioctl_arg_t arg;
    cistpl_vers_1_t *vers = NULL;
    int i, ret, match;
    int has_cis = 0;

    /* Do we have a CIS structure? */
    ret = ioctl(s->fd, DS_VALIDATE_CIS, &arg);
    has_cis = (arg.cisinfo.Chains > 0);
    
    /* Try to read VERS_1 tuple */
    if (has_cis) {
	arg.tuple.TupleData = arg.tuple_parse.data;
	arg.tuple.TupleDataMax = sizeof(arg.tuple_parse.data);
	arg.tuple.Attributes = 0;
	arg.tuple.DesiredTuple = CISTPL_VERS_1;
	do {
	    ret = ioctl(s->fd, DS_GET_FIRST_TUPLE, &arg);
	    if (ret != 0)
		break;
	    arg.tuple.TupleOffset = 0;
	    ret = ioctl(s->fd, DS_GET_TUPLE_DATA, &arg);
	    if (ret != 0) {
		syslog(LOG_INFO, "error reading CIS data on socket %d: %m",
		       ns);
		break;
	    }
	    ret = ioctl(s->fd, DS_PARSE_TUPLE, &arg);
	    if (ret != 0) {
		syslog(LOG_INFO, "error parsing CIS on socket %d: %m", ns);
		break;
	    }
	    vers = &arg.tuple_parse.parse.version_1;
	} while (0);
    }
    
    match = 0;
    for (card = root_card; card; card = card->next) {
	switch (card->ident_type) {

	case BLANK_IDENT:
	    if (!has_cis) match = 1;
	    break;
	    
	case VERS_1_IDENT:
	    if ((vers == NULL) || (card->vers.ns != vers->ns))
		break;
	    for (i = 0; i < vers->ns; i++) {
		if (strcmp(card->vers.pi[i], "*") == 0)
		    continue;
		if (strcmp(card->vers.pi[i], vers->pi[i]) != 0)
		    break;
	    }
	    if (i < vers->ns)
		break;
	    match = 1;
	    break;
	    
	case TUPLE_IDENT:
	    arg.tuple.DesiredTuple = card->tuple.code;
	    ret = ioctl(s->fd, DS_GET_FIRST_TUPLE, &arg);
	    if (ret != 0) break;
	    arg.tuple.TupleOffset = card->tuple.ofs;
	    ret = ioctl(s->fd, DS_GET_TUPLE_DATA, &arg);
	    if (ret != 0) break;
	    if (strncmp(arg.tuple_parse.data, card->tuple.info,
			strlen(card->tuple.info)) != 0)
		break;
	    match = 1;
	    break;
	    
	}
	if (match) break;
    }
    if (match)
	syslog(LOG_INFO, "socket %d: %s", ns, card->name);
    else {
	syslog(LOG_INFO, "unsupported card in socket %d", ns);
	if (vers && (vers->ns > 0)) {
	    char v[256] = "";
	    for (i = 0; i < vers->ns; i++)
		sprintf(v+strlen(v), "%s\"%s\"",
			(i>0) ? ", " : "", vers->pi[i]);
	    syslog(LOG_INFO, "version info: %s", v);
	}
	else
	    syslog(LOG_INFO, "no version info found");
    }
    return card;
} /* lookup_card */

/*====================================================================*/

static int execute(char *msg, char *cmd)
{
    int ret;
    syslog(LOG_INFO, "executing: '%s'", cmd);
    ret = system(cmd);
    if (WIFEXITED(ret)) {
	if (WEXITSTATUS(ret))
	    syslog(LOG_INFO, "%s exited with status %d",
		   msg, WEXITSTATUS(ret));
	return WEXITSTATUS(ret);
    }
    else {
	syslog(LOG_INFO, "%s exited on signal %d",
	       msg, WTERMSIG(ret));
    }
    return -1;
} /* execute */

/*====================================================================*/

static int execute_on_dev(char *msg, char *fmt,
			  int sn, bind_info_t *bind)
{
    int ret;
    char cmd[1024], *s;

    s = cmd;
    while (*fmt != '\0') {
	if ((fmt[0] == '%') && (fmt[1] != '\0') && (fmt[2] == '%')) {
	    fmt++;
	    switch (*fmt) {
	    case 'd':
		strcpy(s, bind->name);
		break;
	    case 'm':
		sprintf(s, "%d", bind->major);
		break;
	    case 'n':
		sprintf(s, "%d", bind->minor);
		break;
	    case 's':
		sprintf(s, "%d", sn);
		break;
	    default:
		*s = '\0';
	    }
	    s += strlen(s);
	    fmt += 2;
	}
	else
	    *s++ = *fmt++;
    }
    *s = '\0';
    
    syslog(LOG_INFO, "executing: '%s'", cmd);
    ret = system(cmd);
    if (WIFEXITED(ret)) {
	if (WEXITSTATUS(ret))
	    syslog(LOG_INFO, "%s exited with status %d",
		   msg, WEXITSTATUS(ret));
	return WEXITSTATUS(ret);
    }
    else {
	syslog(LOG_INFO, "%s exited on signal %d",
	       msg, WTERMSIG(ret));
    }
    return -1;
} /* execute_on_dev */

/*====================================================================*/

static void do_reset(int i)
{
    socket_info_t *s = &socket[i];
    int ret;
    
    if (s->state & (SOCKET_PRESENT|SOCKET_READY)) {
	syslog(LOG_INFO, "bogus insertion event on socket %d", i);
	return;
    }
    s->state |= SOCKET_PRESENT;
    syslog(LOG_INFO, "initializing socket %d", i);
    
    /* Pause 20 ms before resetting */
    usleep(20000);
    ret = ioctl(s->fd, DS_RESET_CARD, NULL);
    if (ret != 0) {
	syslog(LOG_INFO, "reset socket %d failed: %m", i);
	return;
    }
}

/*====================================================================*/

static void do_insert(int sn)
{
    socket_info_t *s = &socket[sn];
    card_info_t *card;
    device_info_t **dev;
    bind_info_t *bind;
    int i, j, ret;
    char cmd[128];

    if (s->state & SOCKET_READY) {
	syslog(LOG_INFO, "bogus reset event on socket %d", sn);
	return;
    }
    s->state |= SOCKET_READY;
    
    /* Pause 20 ms before accessing card */
    usleep(20000);
    s->card = card = lookup_card(sn);
    if (card == NULL)
	return;

    /* Initialize socket bindings */
    dev = card->device; bind = s->bind;
    for (i = 0; i < card->functions; i++) {
	strcpy((char *)bind[i].dev_info, (char *)dev[i]->dev_info);
	bind[i].instance = NULL;
	bind[i].name[0] = '\0';
    }
    
    /* Install kernel modules */
    for (i = 0; i < card->functions; i++) {
	ret = 0;
	for (j = 0; j < dev[i]->modules; j++) {
	    sprintf(cmd, "/sbin/insmod %s/%s.o", modpath, dev[i]->module[j]);
	    ret = execute("insmod", cmd);
	    if ((ret != 0) && (ret != 2)) break;
	}
	if ((ret != 0) && (ret != 2))
	    return;
    }
    
    /* Bind drivers by their dev_info identifiers */
    for (i = 0; i < card->functions; i++) {
	ret = ioctl(s->fd, DS_BIND_REQUEST, &bind[i]);
	if (ret != 0) {
	    syslog(LOG_INFO, "bind '%s' to socket %d failed: %m",
		   (char *)bind[i].dev_info, sn);
	    if (errno != EBUSY) return;
	}
	for (j = 0; j < 10; j++) {
	    ret = ioctl(s->fd, DS_GET_DEVICE_INFO, &bind[i]);
	    if ((ret == 0) || (errno != EAGAIN))
		break;
	    usleep(100000);
	}
	if (ret != 0) {
	    syslog(LOG_INFO, "get dev info on socket %d failed: %m",
		   sn);
	    return;
	}
	if (strlen(bind[i].name) == 0) {
	    syslog(LOG_INFO, "no dev info for socket %d", sn);
	    return;
	}
    }
    
    /* Run "start" commands */
    for (i = 0; i < card->functions; i++) {
	if (dev[i]->start_cmd)
	    execute_on_dev("start cmd", dev[i]->start_cmd, sn, &bind[i]);
    }
} /* do_insert */

/*====================================================================*/

void do_remove(int sn)
{
    socket_info_t *s = &socket[sn];
    card_info_t *card;
    device_info_t **dev;
    bind_info_t *bind;
    int i, j, ret;
    char cmd[128];

    syslog(LOG_INFO, "shutting down socket %d", sn);
    if (!((s->state & SOCKET_PRESENT) &&
	  (s->state & SOCKET_READY))) {
	s->state &= ~(SOCKET_PRESENT | SOCKET_READY);
	return;
    }
    s->state &= ~(SOCKET_PRESENT | SOCKET_READY);
    
    card = s->card;
    if (card == NULL)
	return;
    
    /* Run "stop" commands */
    dev = card->device; bind = s->bind;
    for (i = 0; i < card->functions; i++) {
	if (dev[i]->stop_cmd && (bind[i].name[0] != '\0'))
	    execute_on_dev("stop cmd", dev[i]->stop_cmd, sn, &bind[i]);
    }

    /* unbind driver instances */
    for (i = 0; i < card->functions; i++) {
	if (bind[i].instance != NULL) {
	    ret = ioctl(s->fd, DS_UNBIND_REQUEST, &s->bind[i]);
	    if (ret != 0)
		syslog(LOG_INFO, "unbind '%s' from socket %d failed: %m",
		       (char *)bind[i].dev_info, sn);
	}
    }

    /* remove kernel modules in inverse order */
    for (i = 0; i < card->functions; i++) {
	for (j = dev[i]->modules-1; j >= 0; j--) {
	    sprintf(cmd, "/sbin/rmmod %s", dev[i]->module[j]);
	    execute("rmmod", cmd);
	}
    }
}

/*====================================================================*/

void adjust_resources(int fd)
{
    adjust_list_t *al;
    int ret;
    char tmp[64];
    
    for (al = root_adjust; al; al = al->next) {
	ret = ioctl(fd, DS_ADJUST_RESOURCE_INFO, &al->adj);
	if (ret != 0) {
	    switch (al->adj.Resource) {
	    case RES_MEMORY_RANGE:
		sprintf(tmp, "memory 0x%p-0x%p",
			al->adj.resource.memory.Base,
			al->adj.resource.memory.Base +
			al->adj.resource.memory.Size - 1);
		break;
	    case RES_IO_RANGE:
		sprintf(tmp, "IO ports %#x-%#x",
			al->adj.resource.io.BasePort,
			al->adj.resource.io.BasePort +
			al->adj.resource.io.NumPorts - 1);
		break;
	    case RES_IRQ:
		sprintf(tmp, "irq %lu", al->adj.resource.irq.IRQ);
		break;
	    }
	    syslog(LOG_INFO, "could not adjust resource: %s: %m", tmp);
	}
    }
} /* adjust_resources */
    
/*====================================================================*/

void done(void)
{
    syslog(LOG_INFO, "exiting");
}

void main(int argc, char *argv[])
{
    int optch, errflg, sockets, major, fd;
    int i, ret, event;
    extern char *optarg;
    fd_set fds;
    FILE *f;

    errflg = 0;
    while ((optch = getopt(argc, argv, "c:m:")) != -1) {
	switch (optch) {
	case 'c':
	    config = strdup(optarg); break;
	case 'm':
	    modpath = strdup(optarg); break;
	default:
	    errflg = 1; break;
	}
    }
    if (errflg || (optind < argc)) {
	fprintf(stderr, "usage: %s [-c config]\n", argv[0]);
	exit(EXIT_FAILURE);
    }

#ifdef DEBUG
    openlog("cardmgr", LOG_PID|LOG_PERROR, LOG_DAEMON);
#else
    openlog("cardmgr", LOG_PID|LOG_CONS, LOG_DAEMON);
#endif

#ifndef DEBUG
    if ((ret = fork()) > 0) exit(0);
    if (ret == -1)
	syslog(LOG_INFO, "forking: %m");
    if (setsid() < 0)
	syslog(LOG_INFO, "detaching from tty: %m");
#endif
    
    syslog(LOG_INFO, "starting");
    syslog(LOG_INFO, "%s", version);
    atexit(&done);

    f = fopen(config, "r");
    if (f == NULL) {
	syslog(LOG_INFO, "could not open %s: %m", config);
	exit(EXIT_FAILURE);
    }
    parse_configfile(f);
    fclose(f);

    major = lookup_dev("pcmcia");
    if (major < 0) {
	syslog(LOG_INFO, "no pcmcia driver in /proc/devices");
	exit(EXIT_FAILURE);
    }
    for (i = 0; i < MAX_SOCKS; i++) {
	fd = open_dev(MKDEV(major, i));
	if (fd < 0) break;
	socket[i].fd = fd;
	socket[i].state = 0;
    }
    if ((fd < 0) && (errno != ENODEV)) {
	syslog(LOG_INFO, "open_dev(socket %d) failed: %m", i);
	exit(EXIT_FAILURE);
    }
    sockets = i;
    if (sockets == 0) {
	syslog(LOG_INFO, "no sockets found!");
	exit(EXIT_FAILURE);
    }
    else
	syslog(LOG_INFO, "watching %d sockets", sockets);

    adjust_resources(socket[0].fd);
    
    for (;;) {
	FD_ZERO(&fds);
	for (i = 0; i < sockets; i++)
	    FD_SET(socket[i].fd, &fds);
	
	while ((ret = select(MAX_SOCKS+4, &fds, NULL, NULL, NULL)) < 0) {
	    syslog(LOG_INFO, "select(): %m");
	    exit(EXIT_FAILURE);
	}

	for (i = 0; i < sockets; i++) {
	    if (!FD_ISSET(socket[i].fd, &fds))
		continue;
	    ret = read(socket[i].fd, &event, 4);
	    if ((ret == -1) && (errno != EAGAIN))
		syslog(LOG_INFO, "read(): %m\n");
	    if (ret == 4) {
		if (event & DS_EVENT_CARD_REMOVAL)
		    do_remove(i);
		if (event & DS_EVENT_CARD_INSERTION)
		    do_reset(i);
		if (event & DS_EVENT_RESET_COMPLETE)
		    do_insert(i);
	    }
	}
    } /* repeat */
}
