/*======================================================================

    PCMCIA controller probe

    Written by David Hinds, dhinds@allegro.stanford.edu
    
======================================================================*/

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include <asm/io.h>

#include "i82365.h"
#include "pd67xx.h"
#include "tcic.h"

static const char *version =
    "probe.c v$Revision: 1.5 $: $Date: 1994/08/18 04:39:11 $ (David Hinds)\n";

/*====================================================================*/

static u_char i365_get(u_short sock, u_short reg)
{
    u_short port = I365_PORT(sock);
    u_char val = I365_REG(sock, reg);
    outb_p(val, port); val = inb_p(port+1);
    return val;
}

static void i365_set(u_short sock, u_short reg, u_char data)
{
    u_short port = I365_PORT(sock);
    u_char val = I365_REG(sock, reg);
    outb_p(val, port); outb_p(data, port+1);
}

/*====================================================================*/

int i365_probe(void)
{
    int val, id, sock, done;
    char *name = "i82365sl";

    printf("Intel PCIC probe: ");
    
    id = sock = done = 0;
    ioperm(I365_BASE, 4, 1);
    ioperm(0x80, 1, 1);
    for (; sock < 8; sock++) {
	val = i365_get(sock, I365_IDENT);
	switch (val) {
	case 0x82: case 0x83:
	    name = "i82365sl";
	    id = val & 7;
	    break;
	case 0x84:
	    name = "VLSI 82C146";
	    id = 0;
	    break;
	case 0x88: case 0x89:
	    name = "IBM Clone";
	    id = val & 7;
	    break;
	default:
	    done = 1;
	}
	if (done) break;
	i365_set(sock, I365_MEM(3)+I365_W_OFF, sock);
    }
    
    if (sock == 0) {
	printf("not found.\n");
	return -ENODEV;
    }
    
    /* Check for bogus clones that ignore top bit of index register */
    if ((sock == 4) && (i365_get(0, I365_MEM(3)+I365_W_OFF) == 2))
	sock = 2;
    
    /* Check for Cirrus CL-PD67xx chips */
    i365_set(0, PD67_CHIP_INFO, 0);
    val = i365_get(0, PD67_CHIP_INFO);
    if ((val & PD67_INFO_CHIP_ID) == PD67_INFO_CHIP_ID) {
	val = i365_get(0, PD67_CHIP_INFO);
	if ((val & PD67_INFO_CHIP_ID) == 0) {
	    if (val & PD67_INFO_SLOTS)
		name = "Cirrus CL-PD672x";
	    else {
		name = "Cirrus CL-PD6710";
		sock = 1;
	    }
	    id = 8 - ((val & PD67_INFO_REV) >> 2);
	}
    }

    printf("%s rev %d found, %d sockets\n", name, id, sock);
    return 0;
    
} /* i365_probe */
  
/*====================================================================*/

static u_short tcic_getw(u_long base, u_char reg)
{
    u_short val = inw(base+reg);
    return val;
}

static void tcic_setw(u_long base, u_char reg, u_short data)
{
    outw(data, base+reg);
}

int tcic_probe_at(u_long base)
{
    int i;
    u_short old;
    
    /* Anything there?? */
    for (i = 0; i < 0x10; i += 2)
	if (tcic_getw(base, i) == 0xffff)
	    return -1;

    printf("  at %#3.3lx: ", base); fflush(stdout);
    
    /* Can we set the addr register? */
    old = tcic_getw(base, TCIC_ADDR);
    tcic_setw(base, TCIC_ADDR, 0);
    if (tcic_getw(base, TCIC_ADDR) != 0) {
	tcic_setw(base, TCIC_ADDR, old);
	return -2;
    }
    
    tcic_setw(base, TCIC_ADDR, 0xc3a5);
    if (tcic_getw(base, TCIC_ADDR) != 0xc3a5)
	return -3;

    /* Can we set some of the memory window registers? */
    tcic_setw(base, TCIC_ADDR+2, TCIC_ADR2_INDREG);
    if (tcic_getw(base, TCIC_ADDR+2) != TCIC_ADR2_INDREG)
	return -4;
    
    tcic_setw(base, TCIC_ADDR, TCIC_MWIN(0, 0)+TCIC_MBASE_X);
    tcic_setw(base, TCIC_DATA, 0x03c3);
    if (tcic_getw(base, TCIC_DATA) != 0x03c3)
	return -5;
    
    tcic_setw(base, TCIC_ADDR, TCIC_MWIN(0, 1)+TCIC_MBASE_X);
    tcic_setw(base, TCIC_DATA, 0x0e1e);
    if (tcic_getw(base, TCIC_DATA) != 0x0e1e)
	return -6;
    
    tcic_setw(base, TCIC_ADDR, TCIC_MWIN(0, 0)+TCIC_MBASE_X);
    if (tcic_getw(base, TCIC_DATA) != 0x03c3)
	return -7;

    tcic_setw(base, TCIC_ADDR, TCIC_MWIN(0, 1)+TCIC_MBASE_X);
    if (tcic_getw(base, TCIC_DATA) != 0x0e1e)
	return -8;
    
    return 2;
}

int tcic_probe(u_long base, int expert)
{
    int sock;

    printf("Databook TCIC/2 probe: "); fflush(stdout);
    
    if (expert) {
	printf("\n"); fflush(stdout);
	ioperm(0, 0x400, 1);
	iopl(3);
	sock = 0;
	for (base = 0; base < 0x1000; base += 0x10) {
	    if (base == 0x80) continue;
	    sock = tcic_probe_at(base);
	    if (sock < -1) {
		printf("ret = %d\n", sock); fflush(stdout);
	    }
	    if (sock > 0) break;
	}
    }
    else {
	ioperm(base, 16, 1);
	ioperm(0x80, 1, 1);
	sock = tcic_probe_at(base);
    }
    
    if (sock <= 0) {
	printf("not found.\n");
	return -ENODEV;
    }
    
    printf("%d sockets at %#6lx\n", sock, base);

    return 0;
    
} /* tcic_probe */

/*====================================================================*/

void main(int argc, char *argv[])
{
    int optch, errflg;
    extern char *optarg;
    int verbose = 0, tcic_expert = 0;
    u_long tcic_base = TCIC_BASE;

    
    errflg = 0;
    while ((optch = getopt(argc, argv, "t:vx")) != -1) {
	switch (optch) {
	case 't':
	    tcic_base = strtoul(optarg, NULL, 0); break;
	case 'x':
	    tcic_expert = 1; break;
	case 'v':
	    verbose = 1; break;
	default:
	    errflg = 1; break;
	}
    }
    if (errflg || (optind < argc)) {
	fprintf(stderr, "usage: %s [-t tcic_base] [-v] [-x]\n", argv[0]);
	exit(EXIT_FAILURE);
    }

    if (verbose) printf("%s", version);

    if (tcic_expert) {
	char s[4];
	printf("WARNING: The 'expert' TCIC/2 search can be dangerous,"
	       " and might\ncrash your system.  Confirm (y/n)? ");
	fflush(stdout);
	if (!fgets(s, 4, stdin) || (strcmp(s, "y\n") != 0))
	    exit(EXIT_SUCCESS);
    }
    
    i365_probe();
    tcic_probe(tcic_base, tcic_expert);
}
