/*======================================================================

    PCMCIA device status checker

    Written by David Hinds, dhinds@allegro.stanford.edu
    
======================================================================*/

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <linux/fs.h>

#include "cs_types.h"
#include "cs.h"
#include "cistpl.h"
#include "ds.h"

static const char *version =
    "status.c v$Revision: 1.6 $: $Date: 1994/08/24 20:27:02 $ (David Hinds)\n";

/*====================================================================*/

int lookup_dev(char *name)
{
    FILE *f;
    int n;
    char s[32], t[32];
    
    f = fopen("/proc/devices", "r");
    if (f == NULL)
	return -1;
    while (fgets(s, 32, f) != NULL) {
	if (sscanf(s, "%d %s", &n, t) == 2)
	    if (strcmp(name, t) == 0)
		break;
    }
    fclose(f);
    if (strcmp(name, t) == 0)
	return n;
    else
	return -1;
} /* lookup_dev */

/*====================================================================*/

int open_dev(dev_t dev)
{
    char *fn;
    int fd;
    
    if ((fn = tmpnam(NULL)) == NULL)
	return -1;
    if (mknod(fn, (S_IFCHR|S_IREAD|S_IWRITE), dev) != 0)
	return -1;
    if ((fd = open(fn, O_RDONLY)) < 0)
	return -1;
    if (unlink(fn) != 0) {
	close(fd);
	return -1;
    }
    return fd;
} /* open_dev */

/*====================================================================*/

void print_status(status_t *status)
{
    printf("  Status: ");
    if (status->CardState & CS_EVENT_CARD_DETECT)
	printf("card present");
    else {
	printf("no card\n");
	return;
    }
    if (status->CardState & CS_EVENT_READY_CHANGE)
	printf(", ready");
    if (status->CardState & CS_EVENT_WRITE_PROTECT)
	printf(", write protect");
    if (status->CardState & CS_EVENT_BATTERY_DEAD)
	printf(", battery dead ");
    if (status->CardState & CS_EVENT_BATTERY_LOW)
	printf(", battery low ");
    printf("\n");
} /* print_status */

/*====================================================================*/

void print_config(config_t *config)
{
    printf("  Vcc = %.1f, Vpp1 = %.1f, Vpp2 = %.1f\n",
	   config->Vcc/10.0, config->Vpp1/10.0, config->Vpp2/10.0);
    printf("  Card type is %s\n", (config->IntType == INT_MEMORY)
	   ? "memory-only" : "memory and I/O");
    printf("  Config register base = %#06lx\n",
	   (long)config->ConfigBase);
    
    if (config->CardValues) {
	printf("  ");
	if (config->CardValues & CV_OPTION_VALUE)
	    printf("  Option = %#04x", config->Option);
	if (config->CardValues & CV_STATUS_VALUE)
	    printf("  Status = %#04x", config->Status);
	if (config->CardValues & CV_PIN_REPLACEMENT)
	    printf("  Pin = %#04x", config->Pin);
	if (config->CardValues & CV_COPY_VALUE)
	    printf("  Copy = %#04x", config->Copy);
	printf("\n");
    }

    if (config->AssignedIRQ != 0) {
	printf("  IRQ %ld is ", config->AssignedIRQ);
	switch (config->IRQAttributes & IRQ_TYPE) {
	case IRQ_TYPE_EXCLUSIVE:
	    printf("exclusive"); break;
	case IRQ_TYPE_TIME:
	    printf("time-multiplexed"); break;
	case IRQ_TYPE_DYNAMIC_SHARING:
	    printf("dynamic shared"); break;
	}
	if (config->IRQAttributes & IRQ_PULSE_ALLOCATED)
	    printf(", pulse mode");
	else
	    printf(", level mode");
	printf(", %s\n", (config->Attributes & CONF_ENABLE_IRQ) ?
	       "enabled" : "disabled");
    }
    
    if (config->NumPorts1 > 0) {
	printf("  I/O window 1: %#06x to %#06x", config->BasePort1,
	       config->BasePort1 + config->NumPorts1 - 1);
	if (config->Attributes1 & IO_SHARED)
	    printf(", shared");
	if (config->Attributes1 & IO_FORCE_ALIAS_ACCESS)
	    printf(", force alias");
	if (config->Attributes1 & IO_DATA_PATH_WIDTH)
	    printf(", 16 bit\n");
	else
	    printf(", 8 bit\n");
    }
    if (config->NumPorts2 > 0) {
	printf("  I/O window 2: %#06x to %#06x", config->BasePort2,
	       config->BasePort2 + config->NumPorts2 - 1);
	if (config->Attributes1 & IO_SHARED)
	    printf(", shared");
	if (config->Attributes1 & IO_FORCE_ALIAS_ACCESS)
	    printf(", force alias");
	if (config->Attributes1 & IO_DATA_PATH_WIDTH)
	    printf(", 16 bit\n");
	else
	    printf(", 8 bit\n");
    }
} /* print_config */

/*====================================================================*/

#define MAX_SOCKS 8

void main(int argc, char *argv[])
{
    int i, major, fd, ret;
    status_t status;
    config_t config;

    if ((argc > 1) && (strcmp(argv[1], "-v") == 0))
	printf("%s", version);

    major = lookup_dev("pcmcia");
    if (major < 0) {
	fprintf(stderr, "no pcmcia driver in /proc/devices\n");
	exit(EXIT_FAILURE);
    }
    for (i = 0; i < MAX_SOCKS; i++) {
	fd = open_dev(MKDEV(major, i));
	if (fd < 0) break;
	printf("Socket %d:\n", i);
	ioctl(fd, DS_GET_STATUS, &status);
	print_status(&status);
	ret = ioctl(fd, DS_GET_CONFIGURATION_INFO, &config);
	if (ret == 0)
	    print_config(&config);
    }
}
