/*======================================================================

    PCMCIA card CIS dump

    Written by David Hinds, dhinds@allegro.stanford.edu
    
======================================================================*/

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <linux/fs.h>

#include "cs_types.h"
#include "cs.h"
#include "cistpl.h"
#include "ds.h"

static const char *version =
    "dump_tuples.c v$Revision: 1.9 $: $Date: 1994/09/01 05:58:05 $ (David Hinds)\n";

/*====================================================================*/

int lookup_dev(char *name)
{
    FILE *f;
    int n;
    char s[32], t[32];
    
    f = fopen("/proc/devices", "r");
    if (f == NULL)
	return -1;
    while (fgets(s, 32, f) != NULL) {
	if (sscanf(s, "%d %s", &n, t) == 2)
	    if (strcmp(name, t) == 0)
		break;
    }
    fclose(f);
    if (strcmp(name, t) == 0)
	return n;
    else
	return -1;
} /* lookup_dev */

/*====================================================================*/

int open_dev(dev_t dev)
{
    char *fn;
    int fd;
    
    if ((fn = tmpnam(NULL)) == NULL)
	return -1;
    if (mknod(fn, (S_IFCHR|S_IREAD|S_IWRITE), dev) != 0)
	return -1;
    if ((fd = open(fn, O_RDONLY)) < 0)
	return -1;
    if (unlink(fn) != 0) {
	close(fd);
	return -1;
    }
    return fd;
} /* open_dev */

/*====================================================================*/

void print_tuple(tuple_parse_t *tup)
{
    int i;
    printf("  Tuple: %#2.2x, link: %#04lx, length: %d\n",
	   tup->tuple.TupleCode, tup->tuple.TupleLink,
	   tup->tuple.TupleDataLen);
    for (i = 0; i < tup->tuple.TupleDataLen; i++) {
	if ((i % 16) == 0) printf("    ");
	printf("%2.2x ", (u_char)tup->data[i]);
	if ((i % 16) == 15) printf("\n");
    }
    if ((i % 16) != 0) printf("\n");
} /* print_tuple */


/*====================================================================*/

void print_cftable_entry(cistpl_cftable_entry_t *entry)
{
    int i;
    
    printf("  config entry %#2.2x%s:\n",
	   entry->index, entry->Default ? " (default)" : "");

    if (entry->features & CISTPL_CFTABLE_ENTRY_IO) {
	cistpl_io_t *io = &entry->io;
	printf("  I/O lines = %d", io->flags & CISTPL_IO_LINES_MASK);
	if (io->flags & CISTPL_IO_8BIT) printf(" [8 bit]");
	if (io->flags & CISTPL_IO_16BIT) printf(" [16 bit]");
	if (io->flags & CISTPL_IO_RANGE) printf(" [range]");
	printf("\n");
	for (i = 0; i < io->num; i++)
	    printf("    base = %#4.4lx, length = %#4.4lx\n",
		   io->base[i], io->len[i]);
    }

    if (entry->features & CISTPL_CFTABLE_ENTRY_IRQ) {
	printf("  Interrupt ");
	if (entry->irq.IRQInfo1 & IRQ_INFO2_VALID)
	    printf("mask = %#4.4lx", entry->irq.IRQInfo2);
	else
	    printf("%ld", entry->irq.IRQInfo1 & IRQ_MASK);
	if (entry->irq.IRQInfo1 & IRQ_LEVEL_ID) printf(" [level]");
	if (entry->irq.IRQInfo1 & IRQ_PULSE_ID) printf(" [pulse]");
	if (entry->irq.IRQInfo1 & IRQ_SHARE_ID) printf(" [shared]");
	printf("\n");
    }
    
} /* print_cftable_entry */

/*====================================================================*/

void print_parse(tuple_parse_t *tup)
{
    int i;
    switch (tup->tuple.TupleCode) {
    case CISTPL_VERS_1:
	printf("  version %d.%d\n", tup->parse.version_1.major,
	       tup->parse.version_1.minor);
	for (i = 0; i < tup->parse.version_1.ns; i++)
	    printf("  \"%s\"", tup->parse.version_1.pi[i]);
	printf("\n");
	break;
    case CISTPL_CONFIG:
	printf("  config base = %#4.4lx, mask = %#4.4lx\n",
	       tup->parse.config.base, tup->parse.config.rmask[0]);
	printf("  last config index = %#2.2x\n",
	       tup->parse.config.last_idx);
	break;
    case CISTPL_CFTABLE_ENTRY:
	print_cftable_entry(&tup->parse.cftable_entry);
	break;
    }
} /* print_parse */

/*====================================================================*/

#define MAX_SOCKS 8

void main(int argc, char *argv[])
{
    int i, major, fd, ret;
    ds_ioctl_arg_t arg;

    if ((argc > 1) && (strcmp(argv[1], "-v") == 0))
	printf("%s", version);

    major = lookup_dev("pcmcia");
    if (major < 0) {
	fprintf(stderr, "no pcmcia driver in /proc/devices\n");
	exit(EXIT_FAILURE);
    }
    for (i = 0; i < MAX_SOCKS; i++) {
	fd = open_dev(MKDEV(major, i));
	if (fd < 0) break;
	printf("Socket %d:\n", i);
	arg.tuple.TupleDataMax = sizeof(arg.tuple_parse.data);
	arg.tuple.Attributes = 0;
	arg.tuple.DesiredTuple = RETURN_FIRST_TUPLE;
	arg.tuple.TupleOffset = 0;
	ret = ioctl(fd, DS_GET_FIRST_TUPLE, &arg);
	if (ret != 0) {
	    perror("no first tuple");
	    continue;
	}
	for (;;) {
	    ret = ioctl(fd, DS_GET_TUPLE_DATA, &arg);
	    if (ret != 0) {
		perror("reading tuple data");
		break;
	    }
	    print_tuple(&arg.tuple_parse);
	    ret = ioctl(fd, DS_PARSE_TUPLE, &arg);
	    if (ret != 0) {
		if (errno != ENOSYS) {
		    fflush(stdout);
		    perror("parse error");
		}
	    }
	    else
		print_parse(&arg.tuple_parse);
	    printf("\n");
	    ret = ioctl(fd, DS_GET_NEXT_TUPLE, &arg);
	    if (ret != 0) {
		perror("next tuple");
		break;
	    }
	}
    }
}
