/*======================================================================

    A PCMCIA ethernet driver for the 3com 3c589 card.
    
    Written by David Hinds, dhinds@allegro.stanford.edu

    The network driver code is based on Donald Becker's 3c589 code:
    
    Written 1994 by Donald Becker.
    Copyright 1993 United States Government as represented by the
    Director, National Security Agency.  This software may be used and
    distributed according to the terms of the GNU Public License,
    incorporated herein by reference.
    Donald Becker may be reached at becker@cesdis1.gsfc.nasa.gov
    
======================================================================*/

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/in.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/bitops.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/if_arp.h>
#include <linux/ioport.h>

#ifdef MODULE
#include <linux/module.h>
#include <tools/version.h>
#endif

#include "cs_types.h"
#include "cs.h"
#include "cistpl.h"
#include "ds.h"

/* To minimize the size of the driver source I only define operating
   constants if they are used several times.  You'll need the manual
   if you want to understand driver details. */
/* Offsets from base I/O address. */
#define EL3_DATA	0x00
#define EL3_CMD		0x0e
#define EL3_STATUS	0x0e
#define ID_PORT		0x100
#define	EEPROM_READ	0x80

#define EL3WINDOW(win_num) outw(0x0800+(win_num), ioaddr + EL3_CMD)

/* Register window 1 offsets, the window used in normal operation. */
#define TX_FIFO		0x00
#define RX_FIFO		0x00
#define RX_STATUS 	0x08
#define TX_STATUS 	0x0B
#define TX_FREE		0x0C	/* Remaining free bytes in Tx buffer. */

#define WN4_MEDIA	0x0A	/* Window 4: Various transceiver/media bits. */
#define MEDIA_TP	0x00C0	/* Enable link beat and jabber for 10baseT. */

struct el3_private {
    struct enet_statistics stats;
};

#ifdef PCMCIA_DEBUG
static int pc_debug = PCMCIA_DEBUG;
static char *version =
    "3c589_cs.c v$Revision: 1.12 $: $Date: 1994/09/06 05:07:00 $ (David Hinds)\n";
#endif

#ifdef TC589_DEBUG
static int tc589_debug = TC589_DEBUG;
#else
static int tc589_debug = 3;
#endif

/*====================================================================*/

/* Parameters that can be set with 'insmod' */

/* Special hook for setting if_port when module is loaded */
static int if_port = -1;

/*====================================================================*/

int tc589_probe(struct device *dev);

static void tc589_config(u_long arg);
static void tc589_release(u_long arg);
static int tc589_event(event_t event, int priority,
		       event_callback_args_t *args);

static int tc589_open(struct device *dev);
static int tc589_close(struct device *dev);

static ushort read_eeprom(short ioaddr, int index);
static void tc589_reset(struct device *dev);
static int el3_config(struct device *dev, struct ifmap *map);
static int el3_open(struct device *dev);
static int el3_start_xmit(struct sk_buff *skb, struct device *dev);
static void el3_interrupt(int reg_ptr);
static void update_stats(int addr, struct device *dev);
static struct enet_statistics *el3_get_stats(struct device *dev);
static int el3_rx(struct device *dev);
static int el3_close(struct device *dev);
static void set_multicast_list(struct device *dev, int num_addrs, void *addrs);

static dev_info_t dev_info = "3c589_cs";

static dev_link_t *tc589_attach(void);
static void tc589_detach(dev_link_t *);

static dev_link_t *dev_list = NULL;

/*======================================================================

    If we're linked into the kernel, this gets called from the net
    driver initialization code in Space.c.  It returns failure,
    since all 3c589 devices are dynamically allocated.
    
======================================================================*/

#ifndef MODULE
int tc589_probe(struct device *dev)
{
#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk(version);
#endif
    register_pcmcia_driver(&dev_info, &tc589_attach, &tc589_detach);
    return -1;
}
#endif

/*======================================================================

    We never need to do anything when a tc589 device is "initialized"
    by the net software, because we only register already-found cards.
    
======================================================================*/

static int tc589_init(struct device *dev)
{
    return 0;
}

/*======================================================================

    tc589_attach() creates an "instance" of the driver, allocating
    local data structures for one device.  The device is registered
    with Card Services.

======================================================================*/

static dev_link_t *tc589_attach(void)
{
    client_reg_t client_reg;
    dev_link_t *link;
    struct device *dev;
    int ret;
    
#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("3c589_attach()\n");
#endif

#ifdef MODULE
    MOD_INC_USE_COUNT;
#endif

    /* Create new ethernet device */
    link = kmalloc(sizeof(struct dev_link_t), GFP_KERNEL);
    memset(link, 0, sizeof(struct dev_link_t));
    link->config.function = &tc589_config;
    link->release.function = &tc589_release;
    link->config.data = link->release.data = (u_long)link;
    link->io.NumPorts1 = 16;
    link->io.Attributes1 = IO_DATA_PATH_WIDTH;
    link->io.IOAddrLines = 4;
    link->irq.Attributes = IRQ_TYPE_EXCLUSIVE;
    link->irq.IRQInfo1 = IRQ_INFO2_VALID|IRQ_LEVEL_ID;
    link->irq.IRQInfo2 = 0x86bc;
    link->conf.Attributes = CONF_ENABLE_IRQ;
    link->conf.Vcc = 50;
    link->conf.IntType = INT_MEMORY_AND_IO;
    link->conf.ConfigBase = (caddr_t)0x2000;
    link->conf.ConfigIndex = 1;
    link->conf.Present = PRESENT_OPTION;
    
    dev = kmalloc(sizeof(struct device), GFP_KERNEL);
    memset(dev, 0, sizeof(struct device));
    dev->name = link->dev_name;
    dev->init = &tc589_init;
    dev->open = &tc589_open;
    dev->tbusy = 1;
    link->priv = dev;
    
    /* Register with Card Services */
    link->next = dev_list;
    dev_list = link;
    client_reg.dev_info = &dev_info;
    client_reg.Attributes = INFO_IO_CLIENT | INFO_CARD_SHARE;
    client_reg.EventMask = CS_EVENT_CARD_DETECT;
    client_reg.event_handler = &tc589_event;
    client_reg.Version = 0x0210;
    client_reg.event_callback_args.client_data = link;
    ret = CardServices(RegisterClient, &link->handle, &client_reg);
    if (ret != 0) {
	printk("3c589_cs: RegisterClient failed, ret = %d\n", ret);
	tc589_detach(link);
	return NULL;
    }
    
    return link;
} /* tc589_attach */

/*======================================================================

    This deletes a driver "instance".  The device is de-registered
    with Card Services.  If it has been released, all local data
    structures are freed.  Otherwise, the structures will be freed
    when the device is released.

======================================================================*/

static void tc589_detach(dev_link_t *link)
{
    dev_link_t **linkp;

#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("3c589_detach(0x%p)\n", link);
#endif
    
    /* Locate device structure */
    for (linkp = &dev_list; *linkp; linkp = &(*linkp)->next)
	if (*linkp == link) break;
    if (*linkp == NULL)
	return;

    if (link->state & DEV_CONFIG) {
	printk("3c589_cs: detach postponed, '%s' still locked\n",
	       link->dev_name);
	link->state |= DEV_STALE_LINK;
	return;
    }
    
    if (link->handle)
	CardServices(DeregisterClient, link->handle);
    
    /* Unlink device structure, free bits */
    *linkp = link->next;
    kfree_s(link->priv, sizeof(struct device));
    kfree_s(link, sizeof(struct dev_link_t));
    
#ifdef MODULE
    MOD_DEC_USE_COUNT;
#endif
    
} /* tc589_detach */

/*======================================================================

    tc589_config() is scheduled to run after a CARD_INSERTION event
    is received, to configure the PCMCIA socket, and to make the
    ethernet device available to the system.
    
======================================================================*/

static void tc589_config(u_long arg)
{
    dev_link_t *link = (dev_link_t *)arg;
    client_handle_t handle;
    struct device *dev;
    int i;
    short ioaddr, *phys_addr;
    
    handle = link->handle;
    dev = link->priv;
    phys_addr = (short *)dev->dev_addr;

#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("3c589_config(0x%p)\n", link);
#endif

    /* Configure card */
    link->state |= DEV_CONFIG;

#ifdef MODULE
    MOD_INC_USE_COUNT;
#endif

    do {
	i = CardServices(RequestIO, link->handle, &link->io);
	if (i != CS_SUCCESS) {
	    printk("3c589_cs: RequestIO failed, ret = %d\n", i);
	    break;
	}
	i = CardServices(RequestIRQ, link->handle, &link->irq);
	if (i != CS_SUCCESS) {
	    printk("3c589_cs: RequestIRQ failed, ret = %d\n", i);
	    break;
	}
	i = CardServices(RequestConfiguration, link->handle, &link->conf);
	if (i != CS_SUCCESS) {
	    printk("3c589_cs: RequestConfiguration failed, ret = %d\n", i);
	    break;
	}
	i = request_irq(link->irq.AssignedIRQ, &el3_interrupt,
			0, "3c589_cs");
	if (i != CS_SUCCESS) {
	    printk("3c589_cs: unable to allocate irq %ld\n",
		   link->irq.AssignedIRQ);
	    break;
	}
	dev->irq = link->irq.AssignedIRQ;
	dev->base_addr = link->io.BasePort1;
	dev->tbusy = 0;
	i = register_netdev(dev);
	if (i != 0) {
	    printk("3c589_cs: register_netdev() failed\n");
	    break;
	}
    } while (0);
    
    link->state &= ~DEV_CONFIG_PENDING;
    if (i != 0) {
	tc589_release((u_long)link);
	return;
    }

    ioaddr = dev->base_addr;
    EL3WINDOW(0);

    /* Read the ethernet address and fill in the 3c589 registers.  Unlike
       other EtherLink III adaptors the on-board EEPROM isn't automatically
       read at power-on time.
       We must swap the ethernet address on little-endian machines. */
    for (i = 0; i < 3; i++)
	phys_addr[i] = htons(read_eeprom(ioaddr, i));
    
    /* The address and resource configuration register aren't loaded from
       the EEPROM and *must* be set to 0 and IRQ3 for the PCMCIA version. */
    outw(0x3f00, ioaddr + 8);

    /* The if_port symbol can be set when the module is loaded */
    if (if_port != -1)
	dev->if_port = if_port;
    else
	dev->if_port = read_eeprom(ioaddr, 8) >> 14;
    
    /* This is vital: the transceiver used must be set in the resource
       configuration register.  It took me many hours to discover this. */
    outw(dev->if_port << 14, ioaddr + 6);
    
    {
	char *if_names[] = {"10baseT", "AUI", "undefined", "BNC"};
	printk("%s: 3c589 PCMCIA: %s port, ",
	       dev->name, if_names[dev->if_port]);
    }
    for (i = 0; i < 6; i++)
	printk(" %02X", dev->dev_addr[i]);
    printk("\n");
    
} /* tc589_config */

/*======================================================================

    After a card is removed, tc589_release() will unregister the net
    device, and release the PCMCIA configuration.  If the device is
    still open, this will be postponed until it is closed.
    
======================================================================*/

static void tc589_release(u_long arg)
{
    dev_link_t *link = (dev_link_t *)arg;
    struct device *dev = link->priv;

#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("3c589_release(0x%p)\n", link);
#endif
    
    if (dev->start) {
	printk("3c589_cs: release postponed, '%s' still open\n",
	       link->dev_name);
	link->state |= DEV_STALE_CONFIG;
	return;
    }
    
    if (link->dev_name[0] != '\0')
	unregister_netdev(dev);
    CardServices(ReleaseConfiguration, link->handle);
    CardServices(ReleaseIO, link->handle, &link->io);
    CardServices(ReleaseIRQ, link->handle, &link->irq);
    if (dev->irq != 0) {
	free_irq(dev->irq);
	irq2dev_map[dev->irq] = NULL;
    }
    
    link->state &= ~DEV_CONFIG;
    if (link->state & DEV_STALE_LINK)
	tc589_detach(link);
    
#ifdef MODULE
    MOD_DEC_USE_COUNT;
#endif
} /* tc589_release */

/*======================================================================

    The card status event handler.  Mostly, this schedules other
    stuff to run after an event is received.  A CARD_REMOVAL event
    also sets some flags to discourage the net drivers from trying
    to talk to the card any more.
    
======================================================================*/

static int tc589_event(event_t event, int priority,
		       event_callback_args_t *args)
{
    dev_link_t *link = args->client_data;

#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("3c589_event()\n");
#endif
    
    switch (event) {
#ifdef PCMCIA_DEBUG
    case CS_EVENT_REGISTRATION_COMPLETE:
	if (pc_debug)
	    printk("3c589_cs: registration complete\n");
	break;
#endif
    case CS_EVENT_CARD_REMOVAL:
	link->state &= ~DEV_PRESENT;
	if (link->state & DEV_CONFIG) {
	    ((struct device *)link->priv)->tbusy = 1;
	    link->release.expires = 5;
	    add_timer(&link->release);
	}
	break;
    case CS_EVENT_CARD_INSERTION:
	link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
	link->config.expires = 5;
	add_timer(&link->config);
	break;
    case CS_EVENT_RESET_REQUEST:
	return 1;
	break;
    }
    return 0;
} /* tc589_event */

/*====================================================================*/

static int tc589_open(struct device *dev)
{
    dev_link_t *link;

#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("3c589_open('%s')\n", dev->name);
#endif
    
    for (link = dev_list; link; link = link->next)
	if (link->priv == dev) break;
    if ((link == NULL) || (link->state & DEV_STALE_CONFIG))
	return -ENODEV;
    
#ifdef MODULE
    MOD_INC_USE_COUNT;
#endif
    
    /* Make up a EL3-specific-data structure. */
    dev->priv = kmalloc(sizeof(struct el3_private), GFP_KERNEL);
    memset(dev->priv, 0, sizeof(struct el3_private));
    
    /* The EL3-specific entries in the device structure. */
    dev->hard_start_xmit = &el3_start_xmit;
    dev->set_config = &el3_config;
    dev->stop = &tc589_close;
    dev->get_stats = &el3_get_stats;
    dev->set_multicast_list = &set_multicast_list;

    ether_setup(dev);
    
    return el3_open(dev);
} /* tc589_open */

/*====================================================================*/

static int tc589_close(struct device *dev)
{
    dev_link_t *link;

#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("3c589_close('%s')\n", dev->name);
#endif
    
    for (link = dev_list; link; link = link->next)
	if (link->priv == dev) break;
    if (link == NULL)
	return -ENODEV;
    
    dev->start = 0;
    if (dev->priv != NULL)
	kfree_s(dev->priv, sizeof(struct el3_private));
    if (link->state & DEV_STALE_CONFIG)
	tc589_release((u_long)link);

    el3_close(dev);
    
#ifdef MODULE
    MOD_DEC_USE_COUNT;
#endif
    
    return 0;
} /* tc589_close */

/*====================================================================*/

/* Read a word from the EEPROM using the regular EEPROM access register.
   Assume that we are in register window zero.
 */
static ushort read_eeprom(short ioaddr, int index)
{
    int timer;
    
    outw(EEPROM_READ + index, ioaddr + 10);
    /* Pause for at least 162 us. for the read to take place. */
    for (timer = 0; timer < 162*4 + 400; timer++)
	SLOW_DOWN_IO;
    return inw(ioaddr + 12);
}

/* Reset and restore all of the 3c589 registers. */
static void tc589_reset(struct device *dev)
{
    ushort ioaddr = dev->base_addr;
    int i;
    
    EL3WINDOW(0);
    outw(0x0001, ioaddr + 4);			/* Activate board. */ 
    outw(dev->if_port << 14, ioaddr + 6);	/* Set the xcvr. */
    outw(0x3f00, ioaddr + 8);			/* Set the IRQ line. */
    
    /* Set the station address in window 2. */
    EL3WINDOW(2);
    for (i = 0; i < 6; i++)
	outb(dev->dev_addr[i], ioaddr + i);
    
    if (dev->if_port == 3)
	/* Start the thinnet transceiver. We should really wait 50ms...*/
	outw(0x1000, ioaddr + EL3_CMD);
    else if (dev->if_port == 0) {
	/* 10baseT interface, enabled link beat and jabber check. */
	EL3WINDOW(4);
	outw(inw(ioaddr + WN4_MEDIA) | MEDIA_TP, ioaddr + WN4_MEDIA);
    }
    
    /* Switch to register set 1 for normal use. */
    EL3WINDOW(1);

    outw(0x8005, ioaddr + EL3_CMD); /* Accept b-case and phys addr only. */
    outw(0xA800, ioaddr + EL3_CMD); /* Turn on statistics. */
    outw(0x2000, ioaddr + EL3_CMD); /* Enable the receiver. */
    outw(0x4800, ioaddr + EL3_CMD); /* Enable transmitter. */
    outw(0x78ff, ioaddr + EL3_CMD); /* Allow all status bits to be seen. */
}

static int el3_config(struct device *dev, struct ifmap *map)
{
    if (map->port != (u_char)(-1)) {
	char *if_names[] = {"10baseT", "AUI", "undefined", "BNC"};
	dev->if_port = map->port;
	printk("%s: 3c589 PCMCIA: switched to %s port\n",
	       dev->name, if_names[dev->if_port]);
    }
    return 0;
}

static int el3_open(struct device *dev)
{
    int ioaddr = dev->base_addr;
    
    EL3WINDOW(0);
    if (tc589_debug > 2)
	printk("%s: Opening, IRQ %d status@%x %4.4x.\n", dev->name,
	       dev->irq, ioaddr + EL3_STATUS, inw(ioaddr + EL3_STATUS));
    
    irq2dev_map[dev->irq] = dev;
    
    dev->interrupt = 0;
    dev->tbusy = 0;
    dev->start = 1;
    
    tc589_reset(dev);
    
    /* Switch to register set 1 for normal use. */
    EL3WINDOW(1);
    
    outw(0x7098, ioaddr + EL3_CMD); /* Set interrupt mask. */
    
    if (tc589_debug > 3)
	printk("%s: Opened 3c589  IRQ %d  status %4.4x.\n",
	       dev->name, dev->irq, inw(ioaddr + EL3_STATUS));
    
    return 0;					/* Always succeed */
}

static int el3_start_xmit(struct sk_buff *skb, struct device *dev)
{
    struct el3_private *lp = (struct el3_private *)dev->priv;
    int ioaddr = dev->base_addr;
    
    /* Transmitter timeout, serious problems. */
    if (dev->tbusy) {
	int tickssofar = jiffies - dev->trans_start;
	if (tickssofar < 10)
	    return 1;
	printk("%s: transmit timed out, tx_status %2.2x status %4.4x.\n",
	       dev->name, inb(ioaddr + TX_STATUS), inw(ioaddr + EL3_STATUS));
	dev->trans_start = jiffies;
	/* Issue TX_RESET and TX_START commands. */
	outw(0x5800, ioaddr + EL3_CMD); /* TX_RESET */
	outw(0x4800, ioaddr + EL3_CMD); /* TX_START */
	dev->tbusy = 0;
    }
    
    if (skb == NULL) {
	dev_tint(dev);
	return 0;
    }
    
    if (skb->len <= 0)
	return 0;
    
    if (tc589_debug > 4) {
	printk("%s: el3_start_xmit(lenght = %ld) called, status %4.4x.\n",
	       dev->name, skb->len, inw(ioaddr + EL3_STATUS));
    }
#ifndef final_version
    {	/* Error-checking code, delete for 1.00. */
	ushort status = inw(ioaddr + EL3_STATUS);
	if (status & 0x0003 		/* IRQ line active, missed one. */
	    && inw(ioaddr + EL3_STATUS) & 3) { 			/* Make sure. */
	    printk("%s: Missed interrupt, status then %04x now %04x"
		   "  Tx %2.2x Rx %4.4x.\n", dev->name, status,
		   inw(ioaddr + EL3_STATUS), inb(ioaddr + TX_STATUS),
		   inw(ioaddr + RX_STATUS));
	    if (status & 0x0002) /* Adaptor failure */
		tc589_reset(dev);
	    outw(0x7800, ioaddr + EL3_CMD); /* Fake interrupt trigger. */
	    outw(0x6899, ioaddr + EL3_CMD); /* Ack IRQ */
	    outw(0x78ff, ioaddr + EL3_CMD); /* Set all status bits visible. */
	}
    }
#endif
    
    /* Avoid timer-based retransmission conflicts. */
    if (set_bit(0, (void*)&dev->tbusy) != 0)
	printk("%s: Transmitter access conflict.\n", dev->name);
    else {
	/* Put out the doubleword header... */
	outw(skb->len, ioaddr + TX_FIFO);
	outw(0x00, ioaddr + TX_FIFO);
	/* ... and the packet rounded to a doubleword. */
	outsl(ioaddr + TX_FIFO, skb->data, (skb->len + 3) >> 2);
	
	dev->trans_start = jiffies;
	if (inw(ioaddr + TX_FREE) > 1536) {
	    dev->tbusy=0;
	} else
	    /* Interrupt us when the FIFO has room for max-sized packet. */
	    outw(0x9000 + 1536, ioaddr + EL3_CMD);
    }

    dev_kfree_skb (skb, FREE_WRITE);
    
    /* Clear the Tx status stack. */
    {
	short tx_status;
	int i = 4;
	
	while (--i > 0	&&	(tx_status = inb(ioaddr + TX_STATUS)) > 0) {
	    if (tc589_debug > 5)
		printk("		Tx status %4.4x.\n", tx_status);
	    if (tx_status & 0x38) lp->stats.tx_aborted_errors++;
	    if (tx_status & 0x30) outw(0x5800, ioaddr + EL3_CMD);
	    if (tx_status & 0x3C) outw(0x4800, ioaddr + EL3_CMD);
	    outb(0x00, ioaddr + TX_STATUS); /* Pop the status stack. */
	}
    }
    return 0;
}

/* The EL3 interrupt handler. */
static void el3_interrupt(int reg_ptr)
{
    int irq = -(((struct pt_regs *)reg_ptr)->orig_eax+2);
    struct device *dev = (struct device *)(irq2dev_map[irq]);
    int ioaddr, status;
    int i = 0;
    
    if (dev == NULL) {
	printk ("el3_interrupt(): irq %d for unknown device.\n", irq);
	return;
    }
    
    if (dev->interrupt)
	printk("%s: Re-entering the interrupt handler.\n", dev->name);
    dev->interrupt = 1;
    
    ioaddr = dev->base_addr;
    status = inw(ioaddr + EL3_STATUS);
    
    if (tc589_debug > 4)
	printk("%s: interrupt, status %4.4x.\n", dev->name, status);
    
    do {
	if ((status & 0xe000) != 0x2000)
	    printk("%s: Ejected a live ethercard!\n", dev->name);
	
	if (status & 0x10)
	    el3_rx(dev);
	
	if (status & 0x08) {
	    if (tc589_debug > 5)
		printk("	TX room bit was handled.\n");
	    /* There's room in the FIFO for a full-sized packet. */
	    outw(0x6808, ioaddr + EL3_CMD); /* Ack IRQ */
	    dev->tbusy = 0;
	    mark_bh(NET_BH);
	}
	if (status & 0x80)				/* Statistics full. */
	    update_stats(ioaddr, dev);
	
	if (++i > 10) {
	    printk("%s: Infinite loop in interrupt, status %4.4x.\n",
		   dev->name, status);
	    /* Clear all interrupts we have handled. */
	    outw(0x68FF, ioaddr + EL3_CMD);
	    break;
	}
	/* Acknowledge the IRQ. */
	outw(0x6891, ioaddr + EL3_CMD); /* Ack IRQ */
    } while ((status = inw(ioaddr + EL3_STATUS)) & 0x01);
    
    if (tc589_debug > 4) {
	printk("%s: exiting interrupt, status %4.4x.\n", dev->name,
	       inw(ioaddr + EL3_STATUS));
    }
    
    dev->interrupt = 0;
    return;
}


static struct enet_statistics *el3_get_stats(struct device *dev)
{
    struct el3_private *lp = (struct el3_private *)dev->priv;
    
    sti();
    update_stats(dev->base_addr, dev);
    cli();
    return &lp->stats;
}

/*  Update statistics.  We change to register window 6, so this should be run
    single-threaded if the device is active. This is expected to be a rare
    operation, and it's simpler for the rest of the driver to assume that
    window 1 is always valid rather than use a special window-state variable.
 */
static void update_stats(int ioaddr, struct device *dev)
{
    struct el3_private *lp = (struct el3_private *)dev->priv;
    
    if (tc589_debug > 5)
	printk("   Updating the statistics.\n");
    /* Turn off statistics updates while reading. */
    outw(0xB000, ioaddr + EL3_CMD);
    /* Switch to the stats window, and read everything. */
    EL3WINDOW(6);
    lp->stats.tx_carrier_errors 	+= inb(ioaddr + 0);
    lp->stats.tx_heartbeat_errors	+= inb(ioaddr + 1);
    /* Multiple collisions. */	   	inb(ioaddr + 2);
    lp->stats.collisions			+= inb(ioaddr + 3);
    lp->stats.tx_window_errors		+= inb(ioaddr + 4);
    lp->stats.rx_fifo_errors		+= inb(ioaddr + 5);
    lp->stats.tx_packets			+= inb(ioaddr + 6);
    lp->stats.rx_packets			+= inb(ioaddr + 7);
    /* Tx deferrals */				inb(ioaddr + 8);
    inw(ioaddr + 10);	/* Total Rx and Tx octets. */
    inw(ioaddr + 12);
    
    /* Back to window 1, and turn statistics back on. */
    EL3WINDOW(1);
    outw(0xA800, ioaddr + EL3_CMD);
    return;
}

static int el3_rx(struct device *dev)
{
    struct el3_private *lp = (struct el3_private *)dev->priv;
    int ioaddr = dev->base_addr;
    int boguscnt = 10;
    short rx_status;
    
    if (tc589_debug > 5)
	printk("	   In rx_packet(), status %4.4x, rx_status %4.4x.\n",
	       inw(ioaddr+EL3_STATUS), inw(ioaddr+RX_STATUS));
    while ((rx_status = inw(ioaddr + RX_STATUS)) > 0) {
	if (rx_status & 0x4000) { /* Error, update stats. */
	    short error = rx_status & 0x3800;
	    lp->stats.rx_errors++;
	    switch (error) {
	    case 0x0000:		lp->stats.rx_over_errors++; break;
	    case 0x0800:		lp->stats.rx_length_errors++; break;
	    case 0x1000:		lp->stats.rx_frame_errors++; break;
	    case 0x1800:		lp->stats.rx_length_errors++; break;
	    case 0x2000:		lp->stats.rx_frame_errors++; break;
	    case 0x2800:		lp->stats.rx_crc_errors++; break;
	    }
	}
	if ( (! (rx_status & 0x4000))
	    || ! (rx_status & 0x1000)) { /* Dribble bits are OK. */
	    short pkt_len = rx_status & 0x7ff;
	    struct sk_buff *skb;
	    
	    skb = alloc_skb(pkt_len+3, GFP_ATOMIC);
	    if (tc589_debug > 4)
		printk("	   Receiving packet size %d status %4.4x.\n",
		       pkt_len, rx_status);
	    if (skb != NULL) {
		skb->len = pkt_len;
		skb->dev = dev;
		
		/* 'skb->data' points to the start of sk_buff data area. */
		insl(ioaddr+RX_FIFO, skb->data,
		     (pkt_len + 3) >> 2);
		netif_rx(skb);
		outw(0x4000, ioaddr + EL3_CMD); /* Rx discard */
		continue;
	    } else if (tc589_debug)
		printk("%s: Couldn't allocate a sk_buff of size %d.\n",
		       dev->name, pkt_len);
	}
	lp->stats.rx_dropped++;
	outw(0x4000, ioaddr + EL3_CMD); /* Rx discard */
	while (--boguscnt > 0 && inw(ioaddr + EL3_STATUS) & 0x1000)
	    printk("	Waiting for 3c589 to discard packet, status %x.\n",
		   inw(ioaddr + EL3_STATUS) );
	if (--boguscnt < 0)
	    break;
    }
    
    if (tc589_debug > 5)
	printk("	   Exiting rx_packet(), status %4.4x, rx_status %4.4x.\n",
	       inw(ioaddr+EL3_STATUS), inw(ioaddr+8));
    
    return 0;
}

/* Set or clear the multicast filter for this adaptor.
   num_addrs == -1	Promiscuous mode, receive all packets
   num_addrs == 0	Normal mode, clear multicast list
   num_addrs > 0	Multicast mode, receive normal and MC packets, and do
			best-effort filtering.
 */
static void
set_multicast_list(struct device *dev, int num_addrs, void *addrs)
{
    short ioaddr = dev->base_addr;
    if (num_addrs > 0) {
	outw(0x8007, ioaddr + EL3_CMD);
    } else if (num_addrs < 0) {
	outw(0x8008, ioaddr + EL3_CMD);
    } else
	outw(0x8005, ioaddr + EL3_CMD);
}

static int el3_close(struct device *dev)
{
    int ioaddr = dev->base_addr;
    
    if (tc589_debug > 2)
	printk("%s: Shutting down ethercard.\n", dev->name);
    
    /* Turn off statistics.	 We update lp->stats below. */
    outw(0xB000, ioaddr + EL3_CMD);
    
    /* Disable the receiver and transmitter. */
    outw(0x1800, ioaddr + EL3_CMD);
    outw(0x5000, ioaddr + EL3_CMD);
    
    if (dev->if_port == 3)
	/* Turn off thinnet power. */
	outw(0xb800, ioaddr + EL3_CMD);
    else if (dev->if_port == 0) {
	/* Disable link beat and jabber, if_port may change ere next open(). */
	EL3WINDOW(4);
	outw(inw(ioaddr + WN4_MEDIA) & ~MEDIA_TP, ioaddr + WN4_MEDIA);
    }
    
    /* Switching back to window 0 disables the IRQ. */
    EL3WINDOW(0);
    /* But we explicitly zero the IRQ line select anyway. */
    outw(0x0f00, ioaddr + 8);
        
    /* Check if the card actually exists before updating the statistics. */
    if (inw(ioaddr+EL3_STATUS) != 0xffff)
	update_stats(ioaddr, dev);
    return 0;
}

/*====================================================================*/

#ifdef MODULE
char kernel_version[] = UTS_RELEASE;

int init_module(void)
{
#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk(version);
#endif
    register_pcmcia_driver(&dev_info, &tc589_attach, &tc589_detach);
    return 0;
}

void cleanup_module(void)
{
    printk("3c589_cs: unloading\n");
    unregister_pcmcia_driver(&dev_info);
}
#endif /* MODULE */
