/*======================================================================

    PCMCIA Card Information Structure parser

    Written by David Hinds, dhinds@allegro.stanford.edu
    
======================================================================*/

#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/major.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/sched.h>

#include "cs_types.h"
#include "cs.h"
#include "cistpl.h"

/*======================================================================

    Parsing routines for individual tuples
    
======================================================================*/

#define MIN(a, b) (((a) < (b)) ? (a) : (b))

static int parse_vers_1(tuple_t *tuple, cistpl_vers_1_t *vers_1)
{
    int i, len, sl;
    u_char *p;
    
    len = MIN(tuple->TupleDataLen, tuple->TupleDataMax);
    if (len < 2) return CS_BAD_TUPLE;
    vers_1->major = tuple->TupleData[0];
    vers_1->minor = tuple->TupleData[1];
    len -= 2;
    p = (char *)tuple->TupleData + 2;
    for (i = 0; i < CISTPL_VERS_1_MAX_PROD_STRINGS; i++) {
	if (len < 2) break;
	for (sl = 0; sl < len; sl++)
	    if ((p[sl] == '\0') || (p[sl] == 0xff)) break;
	if (sl >= CISTPL_VERS_1_STRLEN)
	    return CS_BAD_TUPLE;
	strncpy(vers_1->pi[i], p, sl);
	vers_1->pi[i][sl] = '\0';
	len -= sl+1; p += sl+1;
    }
    vers_1->ns = i;
    return CS_SUCCESS;
} /* parse_vers_1 */

/*====================================================================*/

static int parse_config(tuple_t *tuple, cistpl_config_t *config)
{
    int rasz, rmsz, i;
    u_char *p;

    rasz = tuple->TupleData[0] & 0x03;
    rmsz = (tuple->TupleData[0] & 0x3c) >> 2;
    if (tuple->TupleDataLen < rasz+rmsz+4)
	return CS_BAD_TUPLE;
    config->last_idx = tuple->TupleData[1];
    p = tuple->TupleData + 2;
    config->base = 0;
    for (i = 0; i <= rasz; i++)
	config->base += p[i] << (8*i);
    p += rasz+1;
    for (i = 0; i < 4; i++)
	config->rmask[i] = 0;
    for (i = 0; i <= rmsz; i++)
	config->rmask[i>>2] += p[i] << (8*(i%4));
    return CS_SUCCESS;
} /* parse_config */

/*======================================================================

    The following routines are all used to parse the nightmarish
    config table entries.
    
======================================================================*/

static u_char *parse_power(u_char *p, cistpl_power_t *pwr)
{
    int i;
    
    pwr->present = *p; p++;
    for (i = 0; i < 7; i++)
	if (pwr->present & (1<<i)) {
	    pwr->p[i] = *p;
	    while (*p & 0x80) p++;
	    p++;
	}
    return p;
} /* parse_power */

/*====================================================================*/

static u_char *parse_timing(u_char *p, cistpl_timing_t *timing)
{
    timing->scale = *p; p++;
    if ((timing->scale & CISTPL_TIMING_WAIT) != CISTPL_TIMING_WAIT) {
	timing->wait_speed = *p; p++;
    }
    if ((timing->scale & CISTPL_TIMING_RDY) != CISTPL_TIMING_RDY) {
	timing->rdy_speed = *p; p++;
    }
    if ((timing->scale & CISTPL_TIMING_RSV) != CISTPL_TIMING_RSV) {
	timing->rsv_speed = *p; p++;
    }
    return p;
} /* parse_timing */

/*====================================================================*/

static u_char *parse_io(u_char *p, cistpl_io_t *io)
{
    int i, j, bsz, lsz;
    
    io->flags = *p; p++;

    io->num = (*p & 0x0f) + 1;
    bsz = (*p & 0x30) >> 4;
    if (bsz == 3) bsz++;
    lsz = (*p & 0xc0) >> 6;
    if (lsz == 3) lsz++;
    p++;
    
    for (i = 0; i < io->num; i++) {
	io->base[i] = 0;
	io->len[i] = 1;
	for (j = 0; j < bsz; j++, p++)
	    io->base[i] += *p << (j*8);
	for (j = 0; j < lsz; j++, p++)
	    io->len[i] += *p << (j*8);
    }
    return p;
} /* parse_io */

/*====================================================================*/

static u_char *parse_irq(u_char *p, cistpl_irq_t *irq)
{
    irq->IRQInfo1 = *p; p++;
    if (irq->IRQInfo1 & IRQ_INFO2_VALID) {
	irq->IRQInfo2 = (p[1]<<8) + p[0];
	p += 2;
    }
    return p;
} /* parse_irq */

/*====================================================================*/

#define CISTPL_CFTABLE_ENTRY_INDEX_MASK 0x3f
#define CISTPL_CFTABLE_ENTRY_DEFAULT	0x40
#define CISTPL_CFTABLE_ENTRY_INTERFACE	0x80

static int parse_cftable_entry(tuple_t *tuple,
			       cistpl_cftable_entry_t *entry)
{
    int i;
    u_char *p = tuple->TupleData;

    entry->index = *p & CISTPL_CFTABLE_ENTRY_INDEX_MASK;
    entry->Default = (*p & CISTPL_CFTABLE_ENTRY_DEFAULT) ? 1 : 0;
    if (*p & CISTPL_CFTABLE_ENTRY_INTERFACE)
	entry->interface = *(++p);
    else
	entry->interface = 0;

    /* Process optional features */
    entry->features = *(++p); p++;

    for (i = 0; i < (entry->features & CISTPL_CFTABLE_ENTRY_POWER); i++)
	p = parse_power(p, &entry->power[i]);

    /* Timing options */
    if (entry->features & CISTPL_CFTABLE_ENTRY_TIMING)
	p = parse_timing(p, &entry->timing);
    
    /* I/O window options */
    if (entry->features & CISTPL_CFTABLE_ENTRY_IO)
	p = parse_io(p, &entry->io);
    
    /* Interrupt options */
    if (entry->features & CISTPL_CFTABLE_ENTRY_IRQ)
	p = parse_irq(p, &entry->irq);

    return 0;
} /* parse_cftable_entry */

/*====================================================================*/

int parse_tuple(client_handle_t handle, tuple_t *tuple,
		cisparse_t *parse)
{
    int ret = CS_SUCCESS;
    if (tuple->TupleDataLen > tuple->TupleDataMax)
	return CS_BAD_TUPLE;
    switch (tuple->TupleCode) {
    case CISTPL_VERS_1:
	ret = parse_vers_1(tuple, &parse->version_1);
	break;
    case CISTPL_CONFIG:
	ret = parse_config(tuple, &parse->config);
	break;
    case CISTPL_CFTABLE_ENTRY:
	ret = parse_cftable_entry(tuple, &parse->cftable_entry);
	break;
    default:
	ret = CS_UNSUPPORTED_FUNCTION;
	break;
    }
    return ret;
} /* parse_tuple */
